ProtectionState = new(State);

local max = math.max;

-- Update the state using the data of a combat log event.
-- Return SIM_IGNORE if the event was ignorable, SIM_SHIFT if the icons should be shifted and rotation recalculated,
-- or SIM_RECALCULATE if the rotation should be recalculated without shifting the icons
function ProtectionState:OnEventInternal(
	event, timestamp, eventType, hideCaster, 
	srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, dstRaidFlags, 
	p1, p2, p3, s1, s2, s3, s4, s5, s6, s7, 
	spellinfo)

	local t = GetTime();
	local spellID = p1;
	local school = p3;
	
	--print("[prot:OnEvent] |cFFFF00C0" .. eventType .. "|r / " .. (p2 or "?"));

	-- Update the low health status if the player target changes
--	if(eventType == "PLAYER_TARGET_CHANGED") then
--		return checkLowHealth(self);
--	end

	if(eventType == "SPELL_ENERGIZE") then
		-- EG procs a significant amount of time after WoG. We'll re-run the sim after that
		if(spellID == EFFECT_EG) then
			self:SetHolyPower(UnitPower("player", 9), "EG");
			return SIM_RECALCULATE;
		end
		
		-- Unfortunately, sometimes SPELL_ENERGIZE is received after the corresponding SPELL_SUCCESS,
		-- which means we cannot rely on it to provide accurate HoPo readings.
		--self:SetHolyPower(UnitPower("player", 9));
		return SIM_IGNORE;
	end
	
	-- All instants
	if(eventType == "SPELL_CAST_SUCCESS") then
		-- Take target health changes into account
--		local low = checkLowHealth(self);
		local low = SIM_IGNORE;
		
		-- We will always ignore the "holy" part of HotR
		if((spellID == ABILITY_HOR) and (school == 2)) or (spellID == FEEDBACK_HOR_HOLY) then
			return SIM_IGNORE;
		end

		local result = SIM_IGNORE;
		if(spellID ~= ABILITY_HOR) and (spellID ~= ABILITY_CS) and (spellID ~= FEEDBACK_GC_POWER) and (spellID ~= ABILITY_HS) then
			self:SetHolyPower(UnitPower("player", 9), GetSpellInfo(spellID));
		end	
		
		-- Grand Crusader "casts" will update holy power and can result in rotation changes
		if(spellID == FEEDBACK_GC_POWER) then
			return SIM_RECALCULATE;
		end
		
		-- Rotations are not affected by Holy Shield casts and it can be cast at inconvenient times,
		-- so we'll ignore it completely
		if(spellID == ABILITY_HS) then
			return SIM_IGNORE;
		end

		-- We know that CS will always generate 1 HP, therefore we can adjust our value in advance
		if(spellID == ABILITY_CS) or ((spellID == ABILITY_HOR) and (school == 1)) then
			self:SetHolyPower(self:GetHolyPower()+1, "calc: " .. GetSpellInfo(spellID));
			result = SIM_RECALCULATE;
		end
		
		-- HOR and CS have a shared cooldown
		if(spellID == ABILITY_CS) or (spellID == ABILITY_HOR) then
			self:SetCooldown(ABILITY_CS, t + spellinfo:GetCooldown(ABILITY_CS));
			self:SetCooldown(ABILITY_HOR, t + spellinfo:GetCooldown(ABILITY_HOR));
			result = SIM_SHIFT;
		else
			if(spellinfo:IsKnownSpell(spellID)) then
				self:SetCooldown(spellID, t+spellinfo:GetCooldown(spellID));
				result = SIM_SHIFT;
			end
		end;
		
		-- AS while GC is active will generate holy power
		if(spellID == ABILITY_AS) and (self:IsActive(EFFECT_GC, t)) then
			--print(">>>>>>>> AS+GC!!! <<<<<<<<");
			self:SetHolyPower(self:GetHolyPower()+1, GetSpellInfo(spellID));
			result = max(SIM_RECALCULATE, result);
		end

		-- We'll use it to filter out the second HOR
		return max(low, result);
	end
	
	if(eventType == "SPELL_MISSED") then
		--print("[|cFFFF0000missed|r] " .. (spellID and GetSpellInfo(spellID) or "?"));
		-- CS misses invalidate	the assumption that it generates holy power. Revert to the current level
		-- and recalculate.
		if(spellID == ABILITY_CS) then
			self:SetHolyPower(UnitPower("player", 9), "miss: " .. GetSpellInfo(spellID));		-- Preserving the holy power level
			return SIM_RECALCULATE;
		end
		
		-- Same with the physical component of HotR
		if(spellID == ABILITY_HOR) and (school == 1) then
			self:SetHolyPower(UnitPower("player", 9), "miss: " .. GetSpellInfo(spellID));		-- Preserving the holy power level
			return SIM_RECALCULATE;
		end
		
		return SIM_IGNORE;
	end
	
	if(eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_REFRESH") then
		local expiration = adBuffExpiration(spellID);
		self:SetExpiration(spellID, expiration);
		if(spellID == EFFECT_GC) then
			--print("[|cFFFF0000" .. eventType .. "|r] " .. (spellID and GetSpellInfo(spellID) or "?"));
			self:SetCooldown(ABILITY_AS, nil);		-- Grand Crusader refreshes Avenger's Shield cooldown
			return SIM_RECALCULATE;
		end
		
		-- SD may affect our decisions regarding WoG...
		-- ...UNFORTUNATELY, the event associated with SD proc can arrive up to 1.2 seconds later (!),
		-- and
		if(spellID == EFFECT_SD) then
			return SIM_RECALCULATE;
		end

		return SIM_IGNORE;		-- We do NOT want to update our rotation on unimportant buffs
	end
	
	return SIM_IGNORE;
end

-- Apply the given ability to the state object
function ProtectionState:ApplyAbility(id, t, spellinfo)
	if(id == IDLE) then
		return;
	end
	
	if(id == ABILITY_SHOR) then
		self:SetExpiration(EFFECT_SD, nil);										-- ShoR consumes SD...
		self:SetHolyPower(0);													-- ...and HoPo
		--print("[applyAbility] |cFFFF0000" .. id .. "|r (" .. spellinfo:GetCooldown(id) .. ")");
		return;
	end
	
	if(id == ABILITY_INQ) then
		self:SetExpiration(EFFECT_INQ, t + self:GetHolyPower()*4);				-- 4 seconds per HoPo
		self:SetHolyPower(0);													-- INQ consumes HoPo
		--print("[applyAbility] |cFFFF0000" .. id .. "|r (" .. spellinfo:GetCooldown(id) .. ")");
		return;
	end

	if(id == ABILITY_WOG) then
		self:SetCooldown(id, t + spellinfo:GetCooldown(id));
		self:SetHolyPower(0);													-- WOG consumes HoPo
		--print("[applyAbility] |cFFFF0000" .. id .. "|r (" .. spellinfo:GetCooldown(id) .. ")");
		return;
	end

	if(id == ABILITY_CS) or (id == ABILITY_HOR) then
		self:SetCooldown(ABILITY_CS, t + spellinfo:GetCooldown(ABILITY_CS));		-- Shared CS / HOR cooldown
		self:SetCooldown(ABILITY_HOR, t + spellinfo:GetCooldown(ABILITY_HOR));
		self:SetHolyPower(self:GetHolyPower()+1);
		--print("[applyAbility] |cFFFF0000CS/HOR|r: " .. spellinfo:GetCooldown(CS) .. ": HOPO " .. hopo .. " --> " .. self:GetHolyPower());
		return;
	end
	
	if(id == ABILITY_AS) and (self:IsActive(EFFECT_GC, t)) then					-- GC + AS combo (new in patch 4.1)
		self:SetCooldown(id, t + spellinfo:GetCooldown(id));
		self:SetHolyPower(self:GetHolyPower()+1);
		return;
	end
	
	self:SetCooldown(id, t + spellinfo:GetCooldown(id));
	--print("[applyAbility] |cFFFF0000" .. id .. "|r (" .. spellinfo:GetCooldown(id) .. ")");
end
