-- Defines common functionality for all rotation elements

RotationElement = { };

local EMPTY = { };

-- Constructor
function RotationElement:new(id)
	local x = deepCopy(self);
	x.id = id;
	return x;
end

function RotationElement:fail(reason)
	--DEFAULT_CHAT_FRAME:AddMessage("[re:fail] |cFFFFC000" .. GetSpellInfo(self.id) .. "|r: " .. reason);
end

-- Get the icon corresponding to this rotation element's ID
function RotationElement:GetTexture()
	local _, _, t = GetSpellInfo(self.id);
	return t;
end

-- Can we use this ability?
function RotationElement:CanUse(t, state, spellinfo)
	-- Is this ability on cooldown?
	if(state:IsOnCooldown(self.id, t)) then
		self:fail("on cooldown");
		--print("[RE] " .. GetSpellInfo(self.id) .. ":" .. self.id .. " is |cFFFF0000on cooldown|r @ " .. t);
		return false;
	end
	
	--print("[RE] " .. GetSpellInfo(self.id) .. ":" .. self.id .. " is |cFF00FF00NOT on cooldown|r @ " .. t);
	
	-- Do we have enough HoPo if this ability uses it?
	if(spellinfo:UsesHolyPower(self.id)) then
		local required = self.hopo or 3;
		if(state:GetHolyPower() < required) then
			self:fail("hopo(" .. state:GetHolyPower() .. " < " .. required .. ")");
			return false;
		end
	end
	
	-- Check for (de)buffs that have to be up or down
	if(self.buffs) then
		for id,required in pairs(self.buffs) do
			local a;
			if(spellinfo:IsDebuff(id)) then
				-- Debuffs
				a = state:HasDebuff(id, t);
			else
				-- Buffs
				a = state:IsActive(id, t);
			end

			if((a) and (not required)) or ((not a) and (required)) then
				self:fail(GetSpellInfo(id) .. " must " .. (required and "" or "NOT ") .. "be active");
				return false;
			end
		end
	end
	
	-- Check for remaining time on buffs
	if(self.remainingMax) then
		for id,threshold in pairs(self.remainingMax) do
			local tte = state:GetTimeToExpiration(id, t, spellinfo:IsDebuff(id));
			if(tte > threshold) then
				self:fail(GetSpellInfo(id) .. " remaining time (" .. tte .. " > " .. threshold .. ")");
				return false;
			end
		end
	end
	if(self.remainingMin) then
		for id,threshold in pairs(self.remainingMin) do
			local tte = state:GetTimeToExpiration(id, t, spellinfo:IsDebuff(id));
			if(tte < threshold) then
				self:fail(GetSpellInfo(id) .. " remaining time (" .. tte .. " < " .. threshold .. ")");
				return false;
			end
		end
	end
	
	-- Check for debuff stacks
	if(self.stackMax) then
		for id,threshold in pairs(self.stackMax) do
			if(spellinfo:IsDebuff(id)) then
				local _, stack = state:GetDebuff(id, t);
				if(stack > threshold) then
					self:fail(GetSpellInfo(id) .. " stack (" .. stack .. " > " .. threshold .. ")");
					return false;
				end
			end
		end
	end
	if(self.stackMin) then
		for id,threshold in pairs(self.stackMin) do
			if(spellinfo:IsDebuff(id)) then
				local _, stack = state:GetDebuff(id, t);
				if(stack < threshold) then
					self:fail(GetSpellInfo(id) .. " stack (" .. stack .. " < " .. threshold .. ")");
					return false;
				end
			end
		end
	end

	-- Check for mana
	if(self.manaMax) or (self.manaMin) then
		local baseMana = spellinfo:GetBaseMana();
		
		if(self.manaMax) then
			local required = baseMana * self.manaMax / 100.0;
			local m = UnitMana("player");
			if(m > required) then
				self:fail("mana(" .. m .. " > " .. required .. ")");
				return false;
			end
		end
		if(self.manaMin) then
			local required = baseMana * self.manaMin / 100.0;
			local m = UnitMana("player");
			if(m < required) then
				self:fail("mana(" .. m .. " < " .. required .. ")");
				return false;
			end
		end
	end
	
	-- Check for target health requirement
	local threshold = spellinfo:GetHealthThreshold(self.id);
	if(threshold) then
		if(not UnitExists("target")) then
			self:fail("health(no target)");
			return false;
		end
		
		-- Check whether this spell is usable in the first place
		if(not IsUsableSpell(self.id)) then
			self:fail("health(not usable)");
			return false;
		end
	
		local required = 100.0 * UnitHealth("target") / UnitHealthMax("target");
		if(required >= threshold) then
			self:fail("health(" .. required .. " >= " .. threshold .. ")");
			return false;
		end
	end

	return true;
end

-- Check whether this rotation element has any conditions
function RotationElement:HasConditions()
	if(self.hopo) or (self.manaMin) or (self.manaMax) then
		return true;
	end
	
	for k,v in pairs(self.buffs or EMPTY) do
		return true;
	end
	
	for k,v in pairs(self.remainingMax or EMPTY) do
		return true;
	end
	
	for k,v in pairs(self.remainingMin or EMPTY) do
		return true;
	end

	for k,v in pairs(self.stackMin or EMPTY) do
		return true;
	end

	for k,v in pairs(self.stackMax or EMPTY) do
		return true;
	end
	
	return false;	
end

-- Add a buff to our list of buffs (required = true means it should be active,
-- false means it should NOT be active)
function RotationElement:AddBuff(id, required)
	if(self.buffs == nil) then
		self.buffs = { };
	end
	
	self.buffs[id] = required;
end

-- Convert the rotation element into a table that can be stored in a config variable
function RotationElement:Save()
	local info = { };
	
	info.id = self.id;
	info.buffs = self.buffs;
	info.remainingMin = self.remainingMin;
	info.remainingMax = self.remainingMax;
	info.stackMin = self.stackMin;
	info.stackMax = self.stackMax;
	info.manaMin = self.manaMin;
	info.manaMax = self.manaMax;
	info.hopo = self.hopo;

	return info;
end

-- Restore the contents of the rotation element from a table retrieved from a config variable
function RotationElement:Load(info)
	local re = new(RotationElement);
	re.id = info.id;
	re.buffs = info.buffs;
	re.remainingMin = info.remainingMin;
	re.remainingMax = info.remainingMax;
	re.stackMin = info.stackMin;
	re.stackMax = info.stackMax;
	re.manaMin = info.manaMin;
	re.manaMax = info.manaMax;
	re.hopo = info.hopo;
	return re;
end

-- Get the fully qualified code of the rotation element
function RotationElement:GetCode(spellinfo, colored)
	local c = spellinfo:GetCode(self.id);
	
	local q1 = (self.hopo or "");
	
	local q2;
	if(self.manaMin) and (self.manaMax) then
		q2 = "M=" .. self.manaMin .. ".." .. self.manaMax ..  "%";
	elseif(self.manaMin) then
		q2 = "M>" .. self.manaMin .. "%";
	elseif(self.manaMax) then
		q2 = "M<" .. self.manaMax .. "%";
	else
		q2 = "";
	end
	
	-- Buff and debuff requirements
	local effects = spellinfo:GetEffects();
	for id,_ in pairs(effects) do
		local b = (self.buffs or EMPTY)[id];
		if(b ~= nil) then
			q2 = q2 .. (b and " +" or " -") .. spellinfo:GetCode(id);
		end
	end
	
	-- Remaining time
	for id,_ in pairs(effects) do
		local minT = (self.remainingMin or EMPTY)[id];
		local maxT = (self.remainingMax or EMPTY)[id];
		local minS = (self.stackMin or EMPTY)[id];
		local maxS = (self.stackMax or EMPTY)[id];

		if(minT ~= nil) and (maxT ~= nil) then
			q2 = q2 .. " " .. spellinfo:GetCode(id) .. "=" .. minT .. ".." .. maxT;
		elseif(minT ~= nil) then
			q2 = q2 .. " " .. spellinfo:GetCode(id) .. ">" .. minT;
		elseif(maxS ~= nil) then
			q2 = q2 .. " n<" .. maxS;
		elseif(minS ~= nil) then
			q2 = q2 .. " n>" .. minS;
		end
	end	
	
	if(colored) then
		return "|cFFFFB0B0" .. c .. "|r" .. q1 .. ((q2 == "") and "" or ("[" .. string.trim(q2) .. "]"));
	else
		return c .. q1 .. ((q2 == "") and "" or ("[" .. string.trim(q2) .. "]"));
	end
end
