-- Generic rotation code
Rotation = { };

-- Create a new, empty rotation
function Rotation:new()
	local r = deepCopy(Rotation);
	r.list = { };
	return r;
end

-- Used by Rotation:Run() to return simulation results to the addon
local function pack(ability, time, idle)
	local info = { };
	info.ability = ability;
	info.time = time;
	info.idle = idle and true or false;
	
	--print("[pack] " .. ability .. ", " .. time .. ", " .. (idle or "?"));
	
	return info;
end

-- Return the next ability according to the priority list (an array of rotation elements)
function Rotation:GetNext(t, state, spellinfo)
	for i=1,#self.list do
		local re = self.list[i];			-- a RotationElement
		if(re:CanUse(t, state, spellinfo)) then
			return re.id;
		end
	end
	
	return IDLE;
end

-- Calculate the next 'n' abilities
function Rotation:Run(t0, n, state, spellinfo)
	local result = { };
	local stateCopy = deepCopy(state);
	local t = t0 + spellinfo:GCD();
	
	for i=1,n do
		local id = self:GetNext(t, stateCopy, spellinfo);
		result[i] = pack(id, t, id == IDLE);
		stateCopy:ApplyAbility(id, t, spellinfo);
		t = t + spellinfo:GCD();
	end
	
	return result;	
end

-- Print the contents of a simulation list returned by Rotation:Run()
function Rotation:Print(t0, header)
	DEFAULT_CHAT_FRAME:AddMessage("---- " .. (header or "") .. " ----");
	for i=1,#self.list do
		local x = self.list[i];
		local name = x.idle and "[Idle]" or GetSpellInfo(x.ability);
		local t = x.time - t0;		
	
		DEFAULT_CHAT_FRAME:AddMessage("    |cFF00FFFF" .. name .. "|r @ " .. format("%4.3f", t));
	end
end

-- Add a new rotation element to this rotation
function Rotation:add(re)
	self.list[#self.list+1] = re;
end

-- Get the number of elements in the rotation
function Rotation:size()
	return #self.list;
end

-- Get nth rotation element
function Rotation:get(n)
	return self.list[n];
end

-- Set nth rotation element
function Rotation:set(n, re)
	self.list[n] = re;
end

-- Remove nth element from the rotation
function Rotation:remove(n)
	return table.remove(self.list, n);
end

-- Insert an element at position 'n'
function Rotation:insert(n, re)
	return table.insert(self.list, n, re);
end

-- Get the name of the rotation
function Rotation:GetName()
	return self.name or "?";
end

-- Set the name of the rotation
function Rotation:SetName(n)
	self.name = n;
end

-- Check whether the rotation is locked (that is, cannot be changed via the editor)
function Rotation:IsLocked()
	return self.locked;
end

-- Set the "locked" flag
function Rotation:SetLocked(x)
	self.locked = x;
end

-- Convert the rotation into a table that can be stored in a config variable
function Rotation:Save()
	local info = { };
	
	info.name = self.name;
	info.list = { };
	for i=1,#self.list do
		info.list[i] = self.list[i]:Save();
	end

	return info;
end

-- Restore the contents of the rotation from a table retrieved from a config variable
function Rotation:Load(info)
	--print("[R:Load] name='" .. info.name .. "' n=" .. #info.list);
	self.name = info.name;
	self.list = { };
	for i=1,#info.list do
		--print("[R:Load] " .. i .. "/" .. #info.list);
		self.list[i] = RotationElement:Load(info.list[i]);
	end
	--print("[R:Load] Total: " .. #self.list);
end

-- Get the rotation code
function Rotation:GetCode(spellinfo, colored)
	local result = "";
	for i=1,#self.list do
		if(result ~= "") then
			result = result .. " | ";
		end
		result = result .. self.list[i]:GetCode(spellinfo, colored);
	end
	return result;
end
