Spec = {
	["rotations"] = { },
	["single"] = { },
	["aoe"] = { }
};

-- Get the map of all rotations
function Spec:GetRotations()
	return self.rotations;
end

-- Save a rotation with the given name. Return true
-- if the rotation was saved, false if a locked
-- rotation with the same name already exists
function Spec:SaveRotation(r)
	local existing = self.rotations[r:GetName()];
	if(existing) and (existing:IsLocked()) then
		return false;
	end
	
	self.rotations[r:GetName()] = deepCopy(r);
	return true;
end

-- Get a rotation by name
function Spec:GetRotation(name)
	return self.rotations[name];
end

-- Get a rotation by type
function Spec:GetRotationByType(aoe)
	local name = (aoe and self:GetAOERotation() or self:GetSingleRotation());
	name = name or self:GetSingleRotation() or self:GetAOERotation();
	return self:GetRotation(name);
end

-- Delete a rotation by name
function Spec:DeleteRotation(name)
	if(self.locked) then
		return false;
	end
	
	self.rotations[name] = nil;

	if(name == self.singleRotation) then
		self.singleRotation = self:GetDefaultSingleRotation();
	end
	if(name == self.aoeRotation) then
		self.aoeRotation = self:GetDefaultAOERotation();
	end

	return true;
end

-- Get name of the current single-target rotation
function Spec:GetSingleRotation()
	return self.singleRotation;
end

-- Get name of the current AOE rotation
function Spec:GetAOERotation()
	return self.aoeRotation;
end

-- Get name of the default single-target rotation
function Spec:GetDefaultSingleRotation()
	error("Spec:GetDefaultSingleRotation() is not implemented");
end

-- Get name of the default AOE rotation
function Spec:GetDefaultAOERotation()
	error("Spec:GetDefaultAOERotation() is not implemented");
end

-- Set name of the current single-target rotation
function Spec:SetSingleRotation(name)
	if(name == self.aoeRotation) then
		self.aoeRotation = self:GetDefaultAOERotation();
	end
	self.singleRotation = name;
end

-- Set name of the current AOE rotation
function Spec:SetAOERotation(name)
	if(name == self.singleRotation) then
		self.singleRotation = self:GetDefaultSingleRotation();
	end
	self.aoeRotation = name;
end

-- Get the name of the class supported by this spec
function Spec:GetClass()
	error("Spec:GetClass() is not implemented");
end

-- Get the ID of the talent tree supported by this spec
function Spec:GetTalentTree()
	error("Spec:GetTalentTree() is not implemented");
end

-- Get the description of this spec
function Spec:GetDescription()
	error("Spec:GetDescription() is not implemented");
end

-- Get the label of the cost range input boxes
function Spec:GetCostRangeLabel()
	error("Spec:GetCostRangeLabel() is not implemented");
end

-- Get the buff indicating the required stance
function Spec:GetRequiredStanceBuff()
	return nil;		-- stance is not required by default
end

-- Get the config map key
function Spec:GetConfigMap()
	if(AngeleDeiSpecSettings == nil) then
		--print("[spec] Initializing config variable");
		AngeleDeiSpecSettings = { };
	end

	local key = self:GetClass() .. "-" .. self:GetTalentTree();
	--print("[spec] Lookup key: '" .. key .. "'");
	for k,v in pairs(AngeleDeiSpecSettings) do
		--print("[spec] Found key: '" .. k .. "'");
	end

	local config = AngeleDeiSpecSettings[key];
	if(config == nil) then
		--print("[spec] Initializing config map");
		config = { };
		AngeleDeiSpecSettings[key] = config;
	end
	
	return config;	
end

-- Save the data of this spec object to config variables
function Spec:Save()
	local config = self:GetConfigMap();

	-- Save rotations
	config.rotations = { };
	for name,r in pairs(self.rotations) do
		if(not r:IsLocked()) then
			config.rotations[r:GetName()] = r:Save();
		end
	end
	
	-- Save keystrokes
	config.single = deepCopy(self.single) or { };
	config.aoe = deepCopy(self.aoe) or { };
	
	-- Save other
	config.singleRotation = self.singleRotation;
	config.aoeRotation = self.aoeRotation;
end

-- Load the data of this spec object from config variables
function Spec:Load()
	--print("[spec] Loading config info...");
	local config = self:GetConfigMap();

	-- Load rotations
	for name,info in pairs(config.rotations or { }) do
		--print("[spec] Loading rotation '" .. name .. "'");
		local r = Rotation:new();
		r:Load(info);
		self.rotations[r:GetName()] = r;
	end
	
	--for id, text in pairs(config.single) do
	--	print("[single] " .. id .. " = " .. (text or "<n/a>"));
	--end
	--for id, text in pairs(config.aoe) do
	--	print("[aoe] " .. id .. " = " .. (text or "<n/a>"));
	--end
	
	-- Load keystrokes
	self.single = deepCopy(config.single) or { };
	self.aoe = deepCopy(config.aoe) or { };
	
	--for id, text in pairs(self.single) do
	--	print("[self.single] " .. id .. " = " .. (text or "<?>"));
	--end
	--for id, text in pairs(self.aoe) do
	--	print("[self.aoe] " .. id .. " = " .. (text or "<?>"));
	--end
	
	-- Load other
	if(config.singleRotation ~= nil) or (config.aoeRotation ~= nil) then
		-- Don't override default values if neither is in the config
		self.singleRotation = (self.rotations[config.singleRotation] ~= nil) and config.singleRotation or self.singleRotation;
		self.aoeRotation = (self.rotations[config.aoeRotation] ~= nil) and config.aoeRotation or self.aoeRotation;
	end
end

-- Get the keystroke map for the given mode (false = single target, true = AoE)
function Spec:GetKeystrokes(aoe)
	return aoe and self.aoe or self.single;
end

-- Get the buff indicating the required stance
function Spec:GetRequiredStanceBuff()
	return nil;		-- stance is not required by default
end

-- Check whether this spec supports Holy Power
function Spec:SupportsHolyPower()
	return false;
end
