Spellinfo = { };

-- Reset all data of this spellinfo object
function Spellinfo:Reset()
	self.spells = { };
	self.effects = { };
	self.other = { };
	self.playerLevel = 0;
	self.baseMana = 1000;
	
	self:AddSpell(IDLE, 0, "Interface\\Icons\\Spell_Nature_Polymorph_Cow");		-- An empty GCD is not a real ability, but...
end

-- Get the ID and cost of the spell used to measure base mana (base mana percentage, 40 = 40%)
function Spellinfo:GetBaseManaAbility()
	error("Spellinfo:GetBaseManaAbility() is not implemented");
end

-- Get the ID and cast time of the spell used to measure GCD duration
function Spellinfo:GetGCDAbility()
	error("Spellinfo:GetGCDAbility() is not implemented");
end

-- Calculate the base mana. This method is level- and class-independent. It will recalculate the value automatically if
-- the player level
function Spellinfo:GetBaseMana()
	local p = UnitLevel("player");
	if(p ~= self.playerLevel) then
		local spellID, percentage = self:GetBaseManaAbility();
		local _,_,_,cost = GetSpellInfo(spellID);
		self.baseMana = math.ceil(100.0 * cost / percentage);
		self.playerLevel = p;
	end
	
	return self.baseMana;
end

-- Add a spell cooldown
function Spellinfo:AddSpell(spellID, code, cooldown, usesHoPo, targetHealth)
	local name, rank, icon, cost = GetSpellInfo(spellID);

	if(icon == nil) then
		icon = ii;
	end
	
	if(icon == nil) then
		if(spellID == IDLE) then
			icon = "Interface\\Icons\\Spell_Nature_Polymorph_Cow";
		else
			icon = "Interface\\Icons\\Spell_Nature_WispSplodeGreen";
		end		
	end
		
	local info = { };
	info.id = spellID;
	info.code = code;
	info.name = name;
	info.icon = icon;
	info.cost = cost or 0;
	--print("Cost of |cFFFFC000" .. (name or "?") .. "|r is " .. info.cost);
	info.cooldown = cooldown or 0;	
	info.hopo = usesHoPo;
	info.health = targetHealth;
	
	self.spells[spellID] = info;
end

-- Add an effect
function Spellinfo:AddEffect(spellID, code, expiration, debuff, stack, label)
	local name, _, icon = GetSpellInfo(spellID);

	local info = { };
	info.id = spellID;
	info.code = code;
	info.name = name;
	info.icon = nil;
	info.expiration = expiration;
	info.debuff = debuff;
	info.icon = icon;
	info.stack = stack or 1;
	info.label = label;

	self.effects[spellID] = info;
end

-- Add another effect or ability we recognize
function Spellinfo:AddOther(spellID)
	self.other[spellID] = true;
end

-- Check whether this spell ID is something we recognize
function Spellinfo:IsKnownSpell(spellID)
	return (self.spells[spellID] or self.effects[spellID] or self.other[spellID]) and true or false;
end

-- Check whether this ability ID is something we recognize
function Spellinfo:IsKnownAbility(spellID)
	return self.spells[spellID] and true or false;
end

-- Check whether this debuff ID is something we recognize
function Spellinfo:IsKnownDebuff(spellID)
	local info = self.effects[spellID];
	if(info == nil) then
		return false;
	end
	return info.debuff and true or false;
end

-- Get the ability cooldown time
function Spellinfo:GetCooldown(spellID)
	if(self.spells[spellID] == nil) then
		error("Spell " .. (spellID or "?") .. " is undefined");
	end
	return self.spells[spellID].cooldown;
end

-- Get the code of the ability (e.g. "SotR" for "Shield of the Righteous")
function Spellinfo:GetCode(spellID)
	local a = self.spells[spellID];
	if(a) then
		return a.code;
	end
	
	local b = self.effects[spellID];
	if(b) then
		return b.code;
	end
	
	return "?";
end

-- Get the name of the ability
function Spellinfo:GetName(spellID)
	return self.spells[spellID].name;
end

-- Get the ability icon
function Spellinfo:GetIcon(spellID)
	return self.spells[spellID].icon;
end

-- Get the effect expiration time
function Spellinfo:GetExpiration(spellID)
	if(self.effects[spellID] == nil) then
		error("Effect " .. (spellID or "?") .. " is undefined");
	end
	return self.effects[spellID].expiration;
end

-- Check whether this ability requires
function Spellinfo:UsesHolyPower(spellID)
	return self.spells[spellID].hopo;
end

-- Check whether this ability requires
function Spellinfo:GetHealthThreshold(spellID)
	return self.spells[spellID].health;
end

-- Get the cost of this spell (mana, energy, rage etc)
function Spellinfo:GetCost(spellID)
	return self.spells[spellID].cost;
end

-- Get the highest stack of the given buff or debuff
function Spellinfo:GetHighestStack(spellID)
	if(self.effects[spellID] == nil) then
		error("Effect " .. (spellID or "?") .. " is undefined");
	end
	return self.effects[spellID].stack or 1;
end

-- Check whether the given effect is a debuff
function Spellinfo:IsDebuff(spellID)
	if(self.effects[spellID] == nil) then
		error("Effect " .. (spellID or "?") .. " is undefined");
	end
	return self.effects[spellID].debuff and true or false;
end

-- Get the map of spells
function Spellinfo:GetSpells()
	return self.spells;
end

-- Get the map of effects (buffs and debuffs)
function Spellinfo:GetEffects()
	return self.effects;
end

-- Measure the current GCD based on the cast time of a spell
function Spellinfo:MeasureGCD()
	local id, time = self:GetGCDAbility();
	local _,_,_,_,_,_,castTime = GetSpellInfo(id);
	return 1.5 * castTime / (time*1000);
end

-- Get the melee GCD duration in seconds
function Spellinfo:GCD()
	return self:MeasureGCD();
end
