-- Return codes for OnEvent() and OnEventInternal()
SIM_IGNORE = 0;
SIM_RECALCULATE = 1;
SIM_SHIFT = 2;

State = {
	["cooldowns"] = { },
	["effects"] = { },
	["debuffs"] = { },
	["holyPower"] = 0
};

-- Set the cooldown expiration time for the given ability
function State:SetCooldown(spellID, time)
	--DEFAULT_CHAT_FRAME:AddMessage("[|cFFFFC000Angeli Dei / state|r] SetCooldown(" .. (spellID or "?") .. ", " .. (time or "?") .. ")");
	self.cooldowns[spellID] = time;
end

-- Get the cooldown expiration time for the given ability
function State:GetCooldown(spellID)
	return self.cooldowns[spellID];
end

-- Return true if the ability is on cooldown
function State:IsOnCooldown(spellID, time)
	local t = self.cooldowns[spellID];
	return (t) and (time < t);
end

-- Set the expiration time for the given effect
function State:SetExpiration(spellID, time)
	--DEFAULT_CHAT_FRAME:AddMessage("[|cFFFFC000Angeli Dei / state|r] SetExpiration(" .. (spellID or "?") .. ", " .. (time or "?") .. ")");
	self.effects[spellID] = time;
end

-- Get the expiration time for the given effect
function State:GetExpiration(spellID)
	return self.effects[spellID];
end

-- Return true if the effect is active
function State:IsActive(spellID, time)
	local t = self.effects[spellID];
	local x = (t) and (time < t);
	--DEFAULT_CHAT_FRAME:AddMessage("[|cFFFFC000Angeli Dei / state|r] IsActive(" .. (GetSpellInfo(spellID) or "?") .. ", " .. (time or "?") .. "): " .. (x and "true" or "false"));

	return x;
end

-- Get the remaining time to expiration (for a buff or a debuff), or zero if the effect already expired
function State:GetTimeToExpiration(spellID, time, debuff)
	local t = debuff and self:GetDebuff(spellID, time) or self.effects[spellID];
	if(t == nil) or (time >= t) then
		return 0;
	else
		return t - time;
	end
end

-- Get the current holy power
function State:GetHolyPower()
	return self.holyPower;
end;

-- Set the current holy power
function State:SetHolyPower(n, reason)
	self.holyPower = math.min(n, 3);
	--print("SetHolyPower(|cFF40FF80" .. self.holyPower .. "|r" .. (reason and (" - " .. reason) or "") .. ")");
end

-- Get the effective time of the most recent simulation (tLast)
function State:GetEffectiveTLast()
	if(self.tLast == nil) then
		self.tLast = GetTime();
	end
	return self.tLast;
end

-- Update the state using the data of a combat log event. This function should be implemented by specs.
-- Return SIM_IGNORE if the event was ignorable, SIM_SHIFT if the icons should be shifted and rotation recalculated,
-- or SIM_RECALCULATE if the rotation should be recalculated without shifting the icons
function State:OnEventInternal(
	event, timestamp, eventType, hideCaster, 
	srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, dstRaidFlags, 
	p1, p2, p3, s1, s2, s3, s4, s5, s6, s7, 
	spellinfo)
	
	error("Not implemented");
end

-- Adjust the internal clock of the state as though an ability has been used
function State:AdjustTLast(timestamp, result, spellinfo)
	-- Ignored events will not adjust tLast
	if(result == SIM_IGNORE) or (result == SIM_RECALCULATE) then
		return result;
	end
	
	local tLast = self:GetEffectiveTLast();
	local gcd = spellinfo:GCD();
	local adjustedResult = result;	-- (result == SIM_SHIFT) and result or SIM_RECALCULATE;
	
	-- An event that happened shortly after tLast: sim as of tLast, no shift
	if(timestamp <= tLast + gcd*0.5) then
		--print("[AdjustTlast] No change");
		return adjustedResult;
	end
	
	-- An event that happened just before the GCD is up: sim as of the *next* tLast and shift
	if(timestamp <= tLast + gcd) then
		--print("[AdjustTlast] |cFFFF0000Next GCD|r: " .. self.tLast .. " --> " .. (tLast + gcd) .. "; T=" .. timestamp);
		self.tLast = tLast + gcd;
		return adjustedResult;
	end
	
	-- An event that happened after the GCD is up: sim as of the current time and shift
	--print("[AdjustTlast] Current time; T-last: " .. tLast .. ", T: " .. timestamp .. ", gcd: " .. gcd);
	self.tLast = GetTime();
	return adjustedResult;	
end

-- Update the state using the data of a combat log event. This function is called by the main file.
-- It updates the value returned by GetEffectiveTLast().
function State:OnEvent(
	event, timestamp, eventType, hideCaster, 
	srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, dstRaidFlags, 
	p1, p2, p3, s1, s2, s3, s4, s5, s6, s7, 
	spellinfo)

	local result = self:OnEventInternal(
		event, timestamp, eventType, hideCaster, 
		srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, dstRaidFlags, 
		p1, p2, p3, s1, s2, s3, s4, s5, s6, s7, 
		spellinfo)
	
	return self:AdjustTLast(timestamp, result, spellinfo);
end

-- Apply the given ability to the state object
function State:ApplyAbility(id, time, spellinfo)
	error("Not implemented");
end

-- Remove a single debuff from the given unit
function State:RemoveDebuff(spellID)
	self.debuffs[spellID] = nil;
end

-- Check whether the debuff is active on the target
function State:HasDebuff(spellID, time)
	local _, stack = self:GetDebuff(spellID, time);
	--print("[" .. GetSpellInfo(spellID) .. "] HasDebuff(): " .. (stack > 0 and "true" or "false"));
	return stack > 0;
end

-- Get the expiration time of the debuff and the stack size
function State:GetDebuff(spellID, time)
	local b = self.debuffs[spellID];
	if(b == nil) then
		return nil, 0;			-- Nope, don't have it
	end
	
	local t = b.expiration;	
	if(t ~= nil) and (time >= t) then
		return b.expiration, 0;	-- The debuff was there, but it has expired
	end
	
	return b.expiration, (b.stack or 0);
end

-- Apply (a dose of) the debuff to the target. Used by the model.
function State:ApplyDebuff(spellID, expiration, time, spellinfo)
	local _, stack = self:GetDebuff(spellID, time);
	local limit = spellinfo:GetHighestStack(spellID);
	--print("[" .. GetSpellInfo(spellID) .. "] New stack: " .. (stack or "?") .. ", limit: " .. (limit or "?"));
	stack = stack + 1;
	if(stack > limit) then
		stack = limit;
	end
	
	local info = self.debuffs[spellID];
	if(info == nil) then
		info = { };
		self.debuffs[spellID] = info;
	end
	
	info.expiration = expiration;
	info.stack = stack;
end

-- Copy all known debuffs of the player from the target. Called by the target switching code.
function State:CopyAllDebuffs(spellinfo)
	self.debuffs = { };
	if(UnitGUID("target") == nil) then
		return;		-- No target, no debuffs to speak of
	end
	
	for i=1,64 do
		local name,_,_,stack,_,_,expiration,caster,_,_,spellID = UnitDebuff("target", i);
		if(name == nil) then
			return;
		end
		if((caster == "player") and (spellinfo:IsKnownDebuff(spellID))) or (spellID == 91565) then
			if(stack < 1) then
				stack = 1;			-- Non-stackable debuffs have zero as the stack size
			end
			--print("[State:CopyAllDebuffs] Copying |cFFFFC0FF" .. GetSpellInfo(spellID) .. "|r:" .. spellID .. " @ " .. stack);
			self.debuffs[spellID] = {
				["stack"] = stack,
				["expiration"] = expiration
			};
		end
	end
end

-- Copy one debuff from the target. Called by the event processing code.
function State:CopyDebuff(id)
	if(UnitGUID("target") == nil) then
		return;		-- No target
	end
	
	for i=1,64 do
		local name,_,_,stack,_,_,expiration,caster,_,_,spellID = UnitDebuff("target", i);
		if(name == nil) then
			self.debuffs[id] = nil;	-- Couldn't find it
			return;
		end
		if((caster == "player") or (spellID == 91565)) and (spellID == id) then
			if(stack < 1) then
				stack = 1;			-- Non-stackable debuffs have zero as the stack size
			end
			--print("[State:CopyDebuff] Copying |cFFFFC0FF" .. GetSpellInfo(spellID) .. "|r:" .. spellID .. " @ " .. stack);
			self.debuffs[id] = {
				["stack"] = stack,
				["expiration"] = expiration
			};
			return;
		end
	end
	self.debuffs[spellID] = nil;	-- Nope, still couldn't find it
end
